/**
 * Code released under WTFPL [http://sam.zoy.org/wtfpl/]
 * 
 * @author Arnaud Pichery [http://coderpeon.wordpress.com]
 */
package com.coderpeon.comptree
{
	import flash.display.DisplayObject;
	import flash.utils.getQualifiedClassName;
	import mx.core.UIComponent;
	import mx.core.IRawChildrenContainer;
	import mx.core.IChildList;
	import mx.utils.StringUtil;

	public class ComponentTreeItem implements IComponentTreeItem
	{
		private var _displayObject: DisplayObject;
		private var _label: String;
		private var _children: Array;
		private var _childrenComputed: Boolean;
		private var _propertiesComputed: Boolean;
		private var _icon: Class;

		// --- Container icons ---
		[Embed(source="assets/Accordion.png")]
		private var _iconAccordion:Class;

		[Embed(source="assets/ApplicationControlBar.png")]
		private var _iconApplicationControlBar:Class;

		[Embed(source="assets/Box.png")]
		private var _iconBox:Class;

		[Embed(source="assets/Canvas.png")]
		private var _iconCanvas:Class;

		[Embed(source="assets/ControlBar.png")]
		private var _iconControlBar:Class;

		[Embed(source="assets/DividedBox.png")]
		private var _iconDividedBox:Class;

		[Embed(source="assets/Form.png")]
		private var _iconForm:Class;

		[Embed(source="assets/FormHeading.png")]
		private var _iconFormHeading:Class;

		[Embed(source="assets/FormItem.png")]
		private var _iconFormItem:Class;

		[Embed(source="assets/Grid.png")] 
		private var _iconGrid:Class;
		
		[Embed(source="assets/HBox.png")]
		private var _iconHBox:Class;

		[Embed(source="assets/HDividedBox.png")]
		private var _iconHDividedBox:Class;

		[Embed(source="assets/Panel.png")] 
		private var _iconPanel:Class;

		[Embed(source="assets/TabNavigator.png")]
		private var _iconTabNavigator:Class;

		[Embed(source="assets/Tile.png")]
		private var _iconTile:Class;

		[Embed(source="assets/TitleWindow.png")]
		private var _iconTitleWindow:Class;

		[Embed(source="assets/VBox.png")]
		private var _iconVBox:Class;

		[Embed(source="assets/VDividedBox.png")]
		private var _iconVDividedBox:Class;

		[Embed(source="assets/ViewStack.png")]
		private var _iconViewStack:Class;

		[Embed(source="assets/AreaChart.png")]
		private var _iconAreaChart: Class;

		[Embed(source="assets/BarChart.png")]
		private var _iconBarChart: Class;

		[Embed(source="assets/BubbleChart.png")]
		private var _iconBubbleChart: Class;

		[Embed(source="assets/CandlestickChart.png")]
		private var _iconCandlestickChart: Class;

		[Embed(source="assets/ColumnChart.png")]
		private var _iconColumnChart: Class;

		[Embed(source="assets/HLOCChart.png")]
		private var _iconHLOCChart: Class;

		[Embed(source="assets/Legend.png")]
		private var _iconLegend: Class;

		[Embed(source="assets/LineChart.png")]
		private var _iconLineChart: Class;

		[Embed(source="assets/PieChart.png")]
		private var _iconPieChart: Class;

		[Embed(source="assets/PlotChart.png")]
		private var _iconPlotChart: Class;

		[Embed(source="assets/Button.png")]
		private var _iconButton: Class;

		[Embed(source="assets/CheckBox.png")]
		private var _iconCheckBox: Class;

		[Embed(source="assets/ColorPicker.png")]
		private var _iconColorPicker: Class;

		[Embed(source="assets/ComboBox.png")]
		private var _iconComboBox: Class;

		[Embed(source="assets/DataGrid.png")]
		private var _iconDataGrid: Class;

		[Embed(source="assets/DateChooser.png")]
		private var _iconDateChooser: Class;

		[Embed(source="assets/DateField.png")]
		private var _iconDateField: Class;

		[Embed(source="assets/HorizontalList.png")]
		private var _iconHorizontalList: Class;

		[Embed(source="assets/HRule.png")]
		private var _iconHRule: Class;

		[Embed(source="assets/HScrollBar.png")]
		private var _iconHScrollBar: Class;

		[Embed(source="assets/HSlider.png")]
		private var _iconHSlider: Class;

		[Embed(source="assets/Image.png")]
		private var _iconImage: Class;

		[Embed(source="assets/Label.png")]
		private var _iconLabel: Class;

		[Embed(source="assets/LinkBar.png")]
		private var _iconLinkBar: Class;

		[Embed(source="assets/LinkButton.png")]
		private var _iconLinkButton: Class;

		[Embed(source="assets/List.png")]
		private var _iconList: Class;

		[Embed(source="assets/Menu.png")]
		private var _iconMenu: Class;

		[Embed(source="assets/MenuBar.png")]
		private var _iconMenuBar: Class;

		[Embed(source="assets/NumericStepper.png")]
		private var _iconNumericStepper: Class;

		[Embed(source="assets/ProgressBar.png")]
		private var _iconProgressBar: Class;

		[Embed(source="assets/RadioButton.png")]
		private var _iconRadioButton: Class;

		[Embed(source="assets/RadioButtonGroup.png")]
		private var _iconRadioButtonGroup: Class;

		[Embed(source="assets/RichTextEditor.png")]
		private var _iconRichTextEditor: Class;

		[Embed(source="assets/Spacer.png")]
		private var _iconSpacer: Class;

		[Embed(source="assets/SWFLoader.png")]
		private var _iconSWFLoader: Class;

		[Embed(source="assets/TabBar.png")]
		private var _iconTabBar: Class;

		[Embed(source="assets/Text.png")]
		private var _iconText: Class;

		[Embed(source="assets/TextArea.png")]
		private var _iconTextArea: Class;

		[Embed(source="assets/TextInput.png")]
		private var _iconTextInput: Class;

		[Embed(source="assets/TileList.png")]
		private var _iconTileList: Class;

		[Embed(source="assets/Tree.png")]
		private var _iconTree: Class;

		[Embed(source="assets/VideoDisplay.png")]
		private var _iconVideoDisplay: Class;

		[Embed(source="assets/VRule.png")]
		private var _iconVRule: Class;

		[Embed(source="assets/VScrollBar.png")]
		private var _iconVScrollBar: Class;

		[Embed(source="assets/VSlider.png")]
		private var _iconVSlider: Class;

		[Embed(source="assets/Tag.png")]
		private var _iconDefault:Class;

		/**
		 * Creates a new instance of this class.
		 */
		public function ComponentTreeItem(displayObj: DisplayObject) {
			if (displayObj == null)
				throw new Error("displayOjbect argument cannot be null");

			_displayObject = displayObj;
		}

	    /**
	     *  The underlying DisplayObject represented by this item.
	     */
		public function get displayObject(): DisplayObject {
			return _displayObject;
		}

	    /**
	     *  The label of the item.
	     */
		public function get label(): String {
			if (!_propertiesComputed) {
				computeProperties();
			}
			return _label;
		}

	    /**
	     *  The icon for the item.
	     */
		public function get icon(): Class {
			if (!_propertiesComputed) {
				computeProperties();
			}
			return _icon;
		}

	    /**
	     *  The children of this item.
	     */
		public function get children(): Array {
			if (!_childrenComputed) {
				_children = computeChildren();
				_childrenComputed = true;
			}
			return _children;
		}
		
		/**
		 * Computes the label & icon for this item from the
		 * underlying DisplayObject
		 */
		private function computeProperties(): void {
			var className: String = flash.utils.getQualifiedClassName(displayObject);

			// Compute label
			_label = className;
			if (_label.indexOf("::") >= 0) {
				_label = _label.substr(2 + _label.indexOf("::"));
			}
			if (displayObject.name != null) {
				_label += " name=\"" + displayObject.name + "\"";
			}
			_label = StringUtil.substitute("<{0}>", _label);

			// Compute icon
			switch (className) {
				case "mx.containers::Accordion": _icon = _iconAccordion; break;
				case "mx.containers::ApplicationControlBar": _icon = _iconApplicationControlBar; break;
				case "mx.containers::Box": _icon = _iconBox; break;
				case "mx.containers::Canvas": _icon = _iconCanvas; break;
				case "mx.containers::ControlBar": _icon = _iconControlBar; break;
				case "mx.containers::DividedBox": _icon = _iconDividedBox; break;
				case "mx.containers::Form": _icon = _iconForm; break;
				case "mx.containers::FormHeading": _icon = _iconFormHeading; break;
				case "mx.containers::FormItem": _icon = _iconFormItem; break;
				case "mx.containers::Grid.": _icon = _iconGrid; break;
				case "mx.containers::HBox": _icon = _iconHBox; break;
				case "mx.containers::HDividedBox": _icon = _iconHDividedBox; break;
				case "mx.containers::Panel": _icon = _iconPanel; break;
				case "mx.containers::TabNavigator": _icon = _iconTabNavigator; break;
				case "mx.containers::Tile": _icon = _iconTile; break;
				case "mx.containers::TitleWindow": _icon = _iconTitleWindow; break;
				case "mx.containers::VBox": _icon = _iconVBox; break;
				case "mx.containers::VDividedBox": _icon = _iconVDividedBox; break;
				case "mx.containers::ViewStack": _icon = _iconViewStack; break;
				case "mx.charts::AreaChart": _icon = _iconAreaChart; break;
				case "mx.charts::BarChart": _icon = _iconBarChart; break;
				case "mx.charts::BubbleChart": _icon = _iconBubbleChart; break;
				case "mx.charts::CandlestickChart": _icon = _iconCandlestickChart; break;
				case "mx.charts::ColumnChart": _icon = _iconColumnChart; break;
				case "mx.charts::HLOCChart": _icon = _iconHLOCChart; break;
				case "mx.charts::Legend": _icon = _iconLegend; break;
				case "mx.charts::LineChart": _icon = _iconLineChart; break;
				case "mx.charts::PieChart": _icon = _iconPieChart; break;
				case "mx.charts::PlotChart": _icon = _iconPlotChart; break;
				case "mx.controls::Button": _icon = _iconButton; break;
				case "mx.controls::CheckBox": _icon = _iconCheckBox; break;
				case "mx.controls::ColorPicker": _icon = _iconColorPicker; break;
				case "mx.controls::ComboBox": _icon = _iconComboBox; break;
				case "mx.controls::DataGrid": _icon = _iconDataGrid; break;
				case "mx.controls::DateChooser": _icon = _iconDateChooser; break;
				case "mx.controls::DateField": _icon = _iconDateField; break;
				case "mx.controls::HorizontalList": _icon = _iconHorizontalList; break;
				case "mx.controls::HRule": _icon = _iconHRule; break;
				case "mx.controls::HScrollBar": _icon = _iconHScrollBar; break;
				case "mx.controls::HSlider": _icon = _iconHSlider; break;
				case "mx.controls::Image": _icon = _iconImage; break;
				case "mx.controls::Label": _icon = _iconLabel; break;
				case "mx.controls::LinkBar": _icon = _iconLinkBar; break;
				case "mx.controls::LinkButton": _icon = _iconLinkButton; break;
				case "mx.controls::List": _icon = _iconList; break;
				case "mx.controls::Menu": _icon = _iconMenu; break;
				case "mx.controls::MenuBar": _icon = _iconMenuBar; break;
				case "mx.controls::NumericStepper": _icon = _iconNumericStepper; break;
				case "mx.controls::ProgressBar": _icon = _iconProgressBar; break;
				case "mx.controls::RadioButton": _icon = _iconRadioButton; break;
				case "mx.controls::RadioButtonGroup": _icon = _iconRadioButtonGroup; break;
				case "mx.controls::RichTextEditor": _icon = _iconRichTextEditor; break;
				case "mx.controls::Spacer": _icon = _iconSpacer; break;
				case "mx.controls::SWFLoader": _icon = _iconSWFLoader; break;
				case "mx.controls::TabBar": _icon = _iconTabBar; break;
				case "mx.controls::Text": _icon = _iconText; break;
				case "mx.controls::TextArea": _icon = _iconTextArea; break;
				case "mx.controls::TextInput": _icon = _iconTextInput; break;
				case "mx.controls::TileList": _icon = _iconTileList; break;
				case "mx.controls::Tree": _icon = _iconTree; break;
				case "mx.controls::VideoDisplay": _icon = _iconVideoDisplay; break;
				case "mx.controls::VRule": _icon = _iconVRule; break;
				case "mx.controls::VScrollBar": _icon = _iconVScrollBar; break;
				case "mx.controls::VSlider": _icon = _iconVSlider; break;
				default: _icon = _iconDefault; break;
			}
		}
		
		/**
		 * Computes the children of this item.
		 */
		public function computeChildren(): Array {
			var component: UIComponent = displayObject as UIComponent;
			if (component == null)
				return null; // Only UIComponents have children.
		
			var children: Array = new Array();
			
			// Add the "standard" children
			for (var i: int = 0; i < component.numChildren; i++) {
				var child: DisplayObject = component.getChildAt(i);
				
				// Check that this child is not already present in the collection
				if (child != null && !containsChild(children, child)) {
					children.push(new ComponentTreeItem(child));
				}
			}

			// Add the "Chrome" children
			if (component is IRawChildrenContainer) {
				var childList: IChildList = IRawChildrenContainer(component).rawChildren;
				var chromeChildren: Array = new Array();

				// Add the chrome children
				for (var k: int = 0; k < childList.numChildren; k++) {
					var kchild: DisplayObject = childList.getChildAt(k);
					
					// Check that this child is not already present in the collection
					if (!containsChild(children, kchild) && kchild != null) {
						chromeChildren.push(new ComponentTreeItem(kchild));
					}
				}
				
				if (chromeChildren.length > 0) {
					var chromeItem: ComponentTreeChrome = new ComponentTreeChrome(chromeChildren);
					children.push(chromeItem);
				}				
			}
			return (children.length == 0) ? null : children;
		}
		
		/**
		 * Indicates whether the specified DisplayObject is present in the given children collection
		 */
		private static function containsChild(children: Array, displayObj: DisplayObject): Boolean {
			for each (var kid: IComponentTreeItem in children) {
				if (kid is ComponentTreeItem && ComponentTreeItem(kid).displayObject == displayObj) {
					return true;
				}
			}
			return false;
		}
	}
}